# Generated by Django 3.1 on 2022-10-25 12:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('auth_token', '0015_auto_20210308_0919'),
    ]

    operations = [
        migrations.AlterField(
            model_name='mobiledevice',
            name='uuid',
            field=models.CharField(db_index=True, max_length=36, verbose_name='UUID'),
        ),
    ]
